;; Generated package description from poker.el  -*- no-byte-compile: t -*-
(define-package "poker" "0.2.0.20201201.164633" "Texas hold 'em poker" 'nil :url "https://elpa.gnu.org/packages/poker.html" :authors '(("Mario Lang" . "mlang@delysid.org")) :maintainer '("Mario Lang" . "mlang@delysid.org") :keywords '("games"))
